/* @(#)fsyunix.c	15.1.1.1 (ESO-DMD) 01/12/01 17:08:17 */
/*===========================================================================
  Copyright (C) 1995 European Southern Observatory (ESO)
 
  This program is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public 
  License along with this program; if not, write to the Free 
  Software Foundation, Inc., 675 Massachusetts Ave, Cambridge, 
  MA 02139, USA.
 
  Correspondence concerning ESO-MIDAS should be addressed as follows:
	Internet e-mail: midas@eso.org
	Postal address: European Southern Observatory
			Data Management Division 
			Karl-Schwarzschild-Strasse 2
			D 85748 Garching bei Muenchen 
			GERMANY
===========================================================================*/

/*++++++++++++++++++++++++++ Module FSY ++++++++++++++++++++++++++++++++
.TYPE Module
.NAME FSY
.LANGUAGE C
.AUTHOR
  K. Banse                              ESO - Garching
.CATEGORY Image files handling.
.COMMENTS 
 holds FSY_EXTBDF, FSY_MODBDF and all the memory interfaces of ND
.VERSION 
 000118
---------------------------------------------------------------------------*/

#include  <midas_def.h>
#include  <errno.h>
#include  <sys/types.h>   /* to be used in extbdf !!! temporarily */
#include  <sys/stat.h>
#include  <osyparms.h>
#include  <osparms.h>
#include  <fileexts.h>

#ifdef FASTMEM_SET
#include <memoryND.h>		/* for allocation stuff */
#endif

static char   block[BLOCK_SIZE];
static struct stat buf;
static int first = 0;
static long int  filsize;

/*

*/

/* ARGSUSED */

void FSY_EXTBDF(fileid, exalq, nvb, status)
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.PURPOSE  Extend Bulk Data File
.METHOD This routine is used to extend an EXISTING Bulk Data File. The
	file is opened using osdseek and osdwrite; the calling program supplies
	the  file identification and the number of bytes to write.
	(i.e. it is assumed that the file has been opened previously)
	If the file is opened successfully, the initial file size
	(measured in virtual blocks) is saved and the file extended
	by the specified number of blocks; the new file size is then
	calculated and returned to the calling program.
.RETURNS nothing
-------------------------------------------------------------------------*/
int   fileid         /* file identification */;
int   exalq          /* extension allocation quantity */;
int   *nvb           /* number of virtual blocks actually allocated */;
int   *status        /* return status (1 if correct, else 0!)  */;

{
int  jstat;
register int i;
unsigned int nblocksz;



if (first == 0)
   {
   for (i=0; i<BLOCK_SIZE; i++) block[i] = '\0';
   first = 1;
   }


/* the file is supposed to have previously been opened */

nblocksz = BLOCK_SIZE;
*status = 1;				/* default to success */


/* seek to the end of file */
    
jstat = osdseek(fileid,0L,FILE_END);
if (jstat == -1) goto bad_return;
    

/* write exalq * BLOCK_SIZE bytes at this location */

for (i=0; i<exalq; i++)
   {
   jstat = osdwrite(fileid,block,nblocksz);
   if (jstat < BLOCK_SIZE) goto bad_return;
   }

fstat(fileid,&buf);		/* get new file size */
filsize = buf.st_size;
filsize /= BLOCK_SIZE;
*nvb = (int) filsize;
return;


bad_return:
*status = 0;
}

/*
^L
*/

/* ARGSUSED */

void FSY_OPNFIL(name, len, iochan, status)
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.PURPOSE Open Binary File
.METHOD
        This routine is used to open an EXISTING Data File for
        subsequent processing. The filename, logical or physical, is
        supplied by the calling program and the file is opened by
        the osmopen interface facility.
.RETURNS nothing
-------------------------------------------------------------------------*/

char  *name           /* file name */;
int   len             /* length of file name, not used in Unix!  */;
int   *iochan         /* iochannel (file identifier in UNIX) */;
int   *status         /* return status (1 if correct, 0 elsewhere )!**/;
   
{

int  filid;


filid = osmopen(name, READ_WRITE);


/* if no permission for attempted operation, try to open it for read only */

if (filid == -1) filid = osmopen(name,READ);


*iochan = filid;
if (filid < 0)              /* open has failed anyway */
   *status = 0;
else
   *status = 1;

}

/*

*/

int FSY_MODBDF(fileid, nobyt, nvb, status)
int   fileid;		/* file identification */
long int nobyt;		/* no. (position) of byte to be written */
int   *nvb;		/* number of blocks allocated for file*/

{

#ifdef __FreeBSD__
  off_t lseek();
#else
  long int lseek();
#endif 
long int  ret;

char c = ' ';



ret = lseek(fileid,nobyt-1,FILE_START);		/* position to `nobyt-1' */
if (ret == -1L) 
   {
   oserror = errno;
   close(fileid);
   return(-1);
   }

if (write(fileid,&c,1) != 1) 
   {						/* Write last byte */
   oserror = errno;
   close(fileid);
   return(-1);
   }

fstat(fileid,&buf);

filsize = buf.st_size;
filsize /= BLOCK_SIZE;
*nvb = (int) filsize;
return 0;
}

/*

*/

#ifdef FASTMEM_SET


/*+++++++++++++++++++++ Module MEMORY ++++++++++++++++++++++++++++++++++++
.LANGUAGE   ANSI C
.IDENTIFICATION  Module MEMORY.C
.AUTHOR   Nicolas Devillard				951107
	  integrated into Midas env.   Klaus Banse	991022

.Description:	memory handling and swapping routines

 The following module offers functionalities for efficient memory
 handling. The basic goal is to move the upper limit of usable RAM on
 the machine to the amount of available disk space (within limits
 fixed by the OS).
 With this library, it is possible to:

 log all calls to malloc/calloc/free, logging data such as the
 allocated amount of space, the calling file and line number, the
 returned pointer value and the exact data location.
 check calls consistency: warn when a pointer is freed several
 times, or when a block has not been deallocated at the end of the
 program.
 print out current process memory usage.
 allows the use of virtual memory, in an invisible way.
 checks out that temporary files are deleted.
 
--------------------------------------------------------------------------*/


/* #define the following to get zillions of debug messages */
/* #define DEBUG_ALLOC */


#define	ALLOC_CELL_NOALLOC	-1
#define	ALLOC_CELL_RAM		 0
#define ALLOC_CELL_SWAP		 1


/*----------------------------------------------------------------------------
 * 								New types
 *--------------------------------------------------------------------------*/

/* This structure holds all allocated pointers (forward linked list)	*/
typedef struct _MALLOC_CELL_ {
	void 				*	pointer;
	size_t					size;
	char				*	filename ;
	int						lineno ;
	int						flavour ;
	char				*	swapfilename ;
	struct _MALLOC_CELL_*	next ;
	struct _MALLOC_CELL_*	prev ;
} malloc_cell;
		 
/*----------------------------------------------------------------------------
 * 							Global variables
 *--------------------------------------------------------------------------*/

/* This structure holds information about everything allocated so far	*/
static malloc_cell * malloc_table = NULL ;

/*
 * Global variables. Their visibility is limited to this module.
 * Access and modification of their values is done through associated
 * member function calls.
 */
static	size_t	total_allocated_RAM = 0 ;
static	size_t	total_allocated_SWAP = 0 ;
static	int		nswapfiles = 0 ;
static  int		nalloc_ptrs = 0 ;

/* 
 * a register is kept to know how many temporary files have been opened
 * it is only used to generate unique file names
 */
static 	int		file_reg = 0 ;

size_t  MemoryLimitInMegs   =    128 ;
size_t  VMLimitInMegs		=	2048 ;
size_t	VMPageSize			=   1024 ;
char    tmpdirname[1024] = "." ;


#ifdef ENABLE_EXTENDED_MEMORY
#undef malloc
#undef calloc
#undef free
#endif



/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * Function	:	ext_malloc_x()
 * In 		:	size_t	
 * Out 		:	pointer to allocated zone
 * Job		:	allocate memory in RAM or SWAP
 * Notice	:
 *--------------------------------------------------------------------------*/

void *
ext_malloc_x(
	size_t		size,
	char	*	filename,
	int			lineno
)
{
	void	*	pointer ;





	/*
	 * See if we need to allocate swap for the requested size.
	 */
	if (size > VMPageSize) {
		if ((size+total_allocated_RAM)>MemoryLimitInMegs*ONE_MEG) {
                 /*
                   printf("swap_malloc for %d bytes\n",size);
                 */
			pointer = swap_malloc_x(size, filename, lineno) ;
			return pointer ;
		}
	}

	/*
	 * Otherwise allocate in RAM
        printf("ram_malloc for %d bytes\n",size);
	 */
	pointer = ram_malloc(size, filename, lineno) ;
	return pointer ;
}


/*----------------------------------------------------------------------------
 * Function	:	ext_calloc_x()
 * In 		:	size_t	
 * Out 		:	pointer to allocated (and zero-filled) zone
 * Job		:	allocate memory in RAM or SWAP
 * Notice	:
 *--------------------------------------------------------------------------*/

void *
ext_calloc_x(
	size_t		n_elem,
	size_t		size,
	char	*	filename,
	int			lineno
)
{
	void	*	pointer ;

	/*
	 * See if we need to allocate swap for the requested size.
	 */
	if (size > VMPageSize) {
		if ((size+total_allocated_RAM)>MemoryLimitInMegs*ONE_MEG) {
			pointer = swap_malloc_x(size, filename, lineno) ;
			return pointer ;
		}
	}
	/*
	 * Otherwise allocate in RAM
	 */
	pointer = ram_calloc(n_elem, size, filename, lineno) ;
	return pointer ;
}



/*---------------------------------------------------------------------------
 * Function	:	ram_malloc()
 * In 		:	Number of Bytes
 * Out 		:	pointer to allocated zone
 * Job		:	Allocates memory (RAM)
 * Notice	:	returns NULL in case of error
 *--------------------------------------------------------------------------*/

void *
ram_malloc(
	size_t		size,
	char	*	filename,
	int			lineno
)
{
	void		*pointer ;
	malloc_cell	*new_cell ;
	malloc_cell	*curr_cell ;

	pointer = (void*)malloc(size) ;
#ifdef DEBUG_ALLOC
	(void)printf( "-> %p = RAM malloc(%ld) in %s (%d)\n",
				  pointer, size, filename, lineno) ;
	(void)fprintf(stderr,
				  "-> %p = RAM malloc(%ld) in %s (%d)\n",
				  pointer, size, filename, lineno) ;
#endif
	if (pointer == NULL) {
		Error("failed to allocate RAM %ld bytes in %s (%d)\n", 
				size, filename, lineno) ;
		Error("total RAM allocated so far : %ld bytes\n", total_allocated_RAM);
		Error("exiting program now\n") ;
		exit(-2000) ;
	}

	total_allocated_RAM += size ;	
	nalloc_ptrs ++ ;
	new_cell = (malloc_cell*)malloc(sizeof(malloc_cell)) ;
	if (new_cell == (malloc_cell*)NULL) {
		Error("fatal error in memory : life functions terminated\n") ;
		exit(2001) ;
	}
	new_cell->pointer = pointer ;
	new_cell->size = size ;
	new_cell->filename = (char*)malloc(1+(int)strlen(filename)) ;
	if (new_cell->filename == (char*)NULL) {
		Error("fatal error in memory: life functions terminated\n") ;
		exit(2001) ;
	}
	(void)strcpy(new_cell->filename, filename) ;
	new_cell->lineno = lineno ; 
	new_cell->flavour = ALLOC_CELL_RAM ; 
	new_cell->next = NULL ;
	new_cell->prev = NULL ;

	/*
	 * Place new memory cell into the structure
	 */
	
	if (malloc_table == NULL) {
		malloc_table = new_cell ;
	} else {
		curr_cell = malloc_table ;
		while (curr_cell->next != NULL) {
			curr_cell = curr_cell->next ;
		}
		curr_cell->next = new_cell ;
		new_cell->prev  = curr_cell ;
	}
	return(pointer) ;
}

/*---------------------------------------------------------------------------
 * Function :   ram_calloc()
 * In       :   number of elements, size of each element
 * Out      :   pointer to allocated zone
 * Job      :   Allocates memory (RAM)
 * Notice   :   returns NULL in case of error
 *--------------------------------------------------------------------------*/

void *
ram_calloc(
	size_t		n_elem,
	size_t		size,
	char	*	filename,
	int			lineno
)
{
    void        *pointer ;
    malloc_cell  *new_cell ;
    malloc_cell  *curr_cell ;

    pointer = (void*)calloc(n_elem, size) ;
#ifdef DEBUG_ALLOC
	(void)fprintf(stderr,
				  "-> %p = RAM calloc(%ld,%ld) in %s (%d)\n",
				  pointer, n_elem, size, filename, lineno) ;
#endif
    if (pointer == NULL) {
        Error("failed to allocate RAM %ld bytes in %s (%d)\n",
				n_elem * size, filename, lineno) ;
		Error("total RAM allocated so far: %ld bytes\n", total_allocated_RAM);
		Error("exiting program now\n") ;
		exit(-2000) ;
    }
 
    total_allocated_RAM += n_elem * size ;
	nalloc_ptrs ++ ;
	new_cell = (malloc_cell*)malloc(sizeof(malloc_cell)) ;
	if (new_cell == NULL) {
		Error("fatal error in memory: life functions terminated\n") ;
		exit(2001) ;
	}

	new_cell->pointer = pointer ;
	new_cell->size = n_elem * size ;
	new_cell->filename = (char*)malloc(1+(int)strlen(filename)) ;
	if (new_cell->filename == (char*)NULL) {
		Error("fatal error in memory: life functions terminated\n") ;
		exit(2001) ;
	}
	(void)strcpy(new_cell->filename, filename) ;
	new_cell->lineno = lineno ;
	new_cell->flavour = ALLOC_CELL_RAM ;
	new_cell->next = NULL ;
	new_cell->prev = NULL ;

	/*
	 * Place new memory cell into the structure
	 */
	
	if (malloc_table == NULL) {
		malloc_table = new_cell ;
	} else {
		curr_cell = malloc_table ;
		while (curr_cell->next != NULL) {
			curr_cell = curr_cell->next ;
		}
		curr_cell->next = new_cell ;
		new_cell->prev  = curr_cell ;
	}
    return(pointer) ;
}

 
/*---------------------------------------------------------------------------
 * Function	:	ext_free_x()
 * In 		:	Pointer to de-allocate
 * Out 		:	void
 * Job		:	handles pointer free
 * Notice	:
 *--------------------------------------------------------------------------*/

void 
ext_free_x(
	void	*	pointer,
	char	*	filename,
	int			lineno
)
{
	malloc_cell	*	curr_cell ;
	int				known_flag = 0 ;

#ifdef DEBUG_ALLOC
	printf("<- free %p at %s (%d)\n", pointer, filename, lineno) ;
	printf("<- free %p at %s (%d)\n", pointer, filename, lineno) ;
#endif
	if (pointer == NULL) {
/*
		Warning("%s (%d): free requested on NULL pointer\n", filename, lineno);
*/
		return ;
	}

	if (malloc_table == NULL) {
/*
		Warning("%s (%d): free requested on unallocated pointer\n",
				filename, lineno) ;
*/
		return ;
	}

	/*
	 * Look for the requested pointer to free in our list
	 */
	curr_cell = malloc_table ;
	while (curr_cell != NULL) {
		if (curr_cell->pointer == pointer) {
			known_flag = 1 ;
			break ;
		}
		curr_cell = curr_cell->next ;
	}
	if (known_flag) {
		if (curr_cell->flavour == ALLOC_CELL_SWAP) {
			/*
			 * do a swap free: this code is actually a replicate of
			 * swap_free(), it saves a function call and a tree search
			 * to copy/paste it here.
			 */
			total_allocated_SWAP -= curr_cell->size ;
			nalloc_ptrs -- ;
			if (munmap(curr_cell->pointer, curr_cell->size)!=0) {
				Error("failed to munmap pointer %s (%d)\n", filename, lineno);
			}
			if (remove(curr_cell->swapfilename)!=0) {
				Error("failed to remove file %s: errno is %d\n",
						curr_cell->swapfilename, errno) ;
			}
			nswapfiles -- ;
			if ((curr_cell->prev == NULL) && (curr_cell->next == NULL)) {
				malloc_table = NULL ;
			} else {
				if (curr_cell->prev == NULL) {
					/*
					 * First cell becomes malloc_table
					 */
					malloc_table = curr_cell->next ;
					malloc_table->prev = NULL ;
				} else {
					curr_cell->prev->next = curr_cell->next ;
				}
				if (curr_cell->next != NULL) {
					curr_cell->next->prev = curr_cell->prev ;
				}
			}
			free(curr_cell->filename) ;
			free(curr_cell->swapfilename) ;
			free(curr_cell) ;
		} else {
			/*
			 * do a ram free: this code is actually a replicate of
			 * ram_free(), it saves a function call and a tree search
			 * to copy/paste it here.
			 */
			free(curr_cell->pointer) ;
			total_allocated_RAM -= curr_cell->size ;
			nalloc_ptrs -- ;
			if ((curr_cell->prev == NULL) && (curr_cell->next == NULL)) {
				malloc_table = NULL ;
			} else {
				if (curr_cell->prev == NULL) {
					/*
					 * First cell becomes malloc_table
					 */
					malloc_table = curr_cell->next ;
					malloc_table->prev = NULL ;
				} else {
					curr_cell->prev->next = curr_cell->next ;
				}
				if (curr_cell->next != NULL) {
					curr_cell->next->prev = curr_cell->prev ;
				}
			}
			free(curr_cell->filename) ;
			free(curr_cell) ;
		}
	} else {
		Warning("%s (%d) free requested on unallocated pointer\n",
				filename, lineno) ;
	}
	return ;
}


/*---------------------------------------------------------------------------
   Function	:	ram_free()
   In 		:	pointer to free, filename and lineno
   Out 		:	void
   Job		:	frees the RAM associated pointer
   Notice	:
 ---------------------------------------------------------------------------*/

void ram_free(
	void 	*	pointer,
	char	*	filename,
	int			lineno
)
{
	malloc_cell	*	curr_cell ;
	int				known_flag = 0 ;

	if (pointer == NULL) {
		Warning("%s (%d): free requested on NULL pointer\n", filename, lineno);
		return ;
	}

	if (malloc_table == NULL) {
		Warning("%s (%d): free requested on unallocated pointer\n",
				filename, lineno) ;
		return ;
	}

	/*
	 * Look for the requested pointer to free in our list
	 */
	curr_cell = malloc_table ;
	while (curr_cell != NULL) {
		if (curr_cell->pointer == pointer) {
			known_flag = 1 ;
			break ;
		}
		curr_cell = curr_cell->next ;
	}
	if (known_flag) {
		if (curr_cell->flavour != ALLOC_CELL_RAM) {
			Error("trying to RAM free a non-RAM pointer: skipping free\n") ;
			return ;
		} else {
			free(curr_cell->pointer) ;
			total_allocated_RAM -= curr_cell->size ;
			nalloc_ptrs -- ;
			if ((curr_cell->prev == NULL) && (curr_cell->next == NULL)) {
				malloc_table = NULL ;
			} else {
				if (curr_cell->prev == NULL) {
					/*
					 * First cell becomes malloc_table
					 */
					malloc_table = curr_cell->next ;
					malloc_table->prev = NULL ;
				} else {
					curr_cell->prev->next = curr_cell->next ;
				}
				if (curr_cell->next != NULL) {
					curr_cell->next->prev = curr_cell->prev ;
				}
			}
			free(curr_cell->filename) ;
			free(curr_cell) ;
		}
	} else {
		Warning("%s (%d) free requested on unallocated pointer\n",
				filename, lineno) ;
	}
	return ;
}



/*---------------------------------------------------------------------------
   Function	:	swap_malloc_x()
   In 		:	Number of bytes to allocate in swap area
   Out 		:	pointer to allocated zone
   Job		:	Allocates memory in swap space, returns NULL if cannot
   Notice	:
   				memory should be freed with swap_free_monit()
				this function should not be called directly but through
				the swap_malloc() macro
 ---------------------------------------------------------------------------*/

void *
swap_malloc_x(
	size_t		size,
	char	*	filename,
	int			lineno
)
{
	void		*	pointer ;
	malloc_cell	*	new_cell ;
	malloc_cell	*	curr_cell ;

	int			 	swapfiled ;
	char		*	fname ;
	char		*	wbuf ;
	int				nbufs ;
	int				i ;

	if (size+total_allocated_SWAP > VMLimitInMegs*ONE_MEG) {
		Error("fatal: swap space overflow --> killing process\n") ;
		exit(-2000) ;
	}

	wbuf = malloc(VMPageSize) ;
	if (wbuf == NULL) {
		Error("fatal: cannot allocate internal buffer\n") ;
		Error("exiting program now\n") ;
		exit(-2000) ;
	}

	/* create swap file with rights: rwxrwxrwx */
	fname = get_tempfile_name() ;
	swapfiled = open(fname, O_RDWR | O_CREAT) ;
	fchmod(swapfiled, S_IRWXU | S_IRWXG | S_IRWXO) ;

    if (size % VMPageSize == 0) {
        nbufs = size / VMPageSize ;
    } else {
        nbufs = 1 + (size / VMPageSize) ;
    }
    for (i=0 ; i<nbufs ; i++) {
        if (write(swapfiled, wbuf, VMPageSize) == -1) {
			Error("fatal: cannot write to swapfile: [%s] full?\n", tmpdirname);
			Error("exiting program now\n") ;
			free(wbuf) ;
			close(swapfiled) ;
			remove(fname) ;
			free(fname) ;
			exit(-2000) ;
		}
    }
	free(wbuf) ;

    /* mmap() the swap file */
    pointer = (void*)mmap(0,
                          size,
                          PROT_READ | PROT_WRITE,
                          MAP_SHARED,
                          swapfiled,
                          0) ;
    if (pointer == (void*)-1) {
        Error("mmap failed with errno = %d\n", errno) ;
		Error("failed to allocate SWAP %ld bytes in %s (%d)\n", 
				size, filename, lineno) ;
		Error("total SWAP allocated so far : %ld bytes\n",total_allocated_SWAP);
		Error("%d swap files opened so far in %s\n", nswapfiles, tmpdirname) ;
        close(swapfiled) ;
		Error("exiting program now\n") ;
		exit(-2000) ;
    }
#ifdef DEBUG_ALLOC
	(void)fprintf(stderr,
				  "-> %p = SWAP malloc(%ld) by %s (%d) in %s/%s\n",
				  pointer, size, filename, lineno, tmpdirname, fname) ;
#endif

	total_allocated_SWAP += size ;	
	nalloc_ptrs ++ ;
	nswapfiles++ ;

	new_cell = (malloc_cell*)malloc(sizeof(malloc_cell)) ;
	if (new_cell == (malloc_cell*)NULL) {
		Error("fatal error in memory: life functions terminated\n") ;
		exit(2001) ;
	}
	new_cell->pointer = pointer ;
	new_cell->size = size ;
	new_cell->filename = (char*)malloc(1+(int)strlen(filename)) ;
	if (new_cell->filename == (char*)NULL) {
		Error("fatal error in memory: life functions terminated\n") ;
		exit(2001) ;
	}
	(void)strcpy(new_cell->filename, filename) ;
	new_cell->lineno = lineno ; 
	new_cell->flavour = ALLOC_CELL_SWAP ; 
	new_cell->swapfilename = fname ;
	new_cell->next = NULL ;
	new_cell->prev = NULL ;
		
	if (malloc_table == NULL) {
		malloc_table = new_cell ;
	} else {
		curr_cell = malloc_table ;
		while (curr_cell->next != NULL) {
			curr_cell = curr_cell->next ;
		}
		curr_cell->next = new_cell ;
		new_cell->prev  = curr_cell ;
	}
	return(pointer) ;
}

/*---------------------------------------------------------------------------
 * Function	:	swap_free_x()
 * In 		:	SWAP pointer to de-allocate
 * Out 		:	void
 * Job		:	handles swap pointer free
 * Notice	:	don't even try to understand...
 *--------------------------------------------------------------------------*/

void 
swap_free_x(
	void	*	pointer,
	char	*	filename,
	int			lineno
)
{
	malloc_cell	*	curr_cell ;
	int				known_flag = 0 ;

	if (pointer == NULL) {
		Warning("%s (%d): free requested on NULL pointer\n", filename, lineno);
		return ;
	}

	if (malloc_table == NULL) {
		Warning("%s (%d): free requested on unallocated pointer\n",
				filename, lineno) ;
		return ;
	}

	/*
	 * Look for the requested pointer to free in our list
	 */
	curr_cell = malloc_table ;
	while (curr_cell != NULL) {
		if (curr_cell->pointer == pointer) {
			known_flag = 1 ;
			break ;
		}
	}
	if (known_flag) {
		if (curr_cell->flavour != ALLOC_CELL_SWAP) {
			Error("trying to RAM free a non-SWAP pointer: skipping free\n") ;
			return ;
		} else {
			total_allocated_SWAP -= curr_cell->size ;
			nalloc_ptrs -- ;
			if (munmap(curr_cell->pointer, curr_cell->size)!=0) {
				Error("failed to munmap pointer %s (%d)\n", filename, lineno);
			}
			if (remove(curr_cell->swapfilename)!=0) {
				Error("failed to remove file %s: errno is %d\n",
						curr_cell->swapfilename, errno) ;
			}
			nswapfiles -- ;
			if ((curr_cell->prev == NULL) && (curr_cell->next == NULL)) {
				malloc_table = NULL ;
			} else {
				if (curr_cell->prev == NULL) {
					/*
					 * First cell becomes malloc_table
					 */
					malloc_table = curr_cell->next ;
					malloc_table->prev = NULL ;
				} else {
					curr_cell->prev->next = curr_cell->next ;
				}
				if (curr_cell->next != NULL) {
					curr_cell->next->prev = curr_cell->prev ;
				}
			}
			free(curr_cell->filename) ;
			free(curr_cell->swapfilename) ;
			free(curr_cell) ;
		}
	} else {
		Warning("%s (%d) free requested on unallocated pointer\n",
				filename, lineno) ;
	}
	return ;
}



/*----------------------------------------------------------------------------
 * Function	:	get_memory_parameter()
 * In 		:	allocated character string	
 * Out 		:	size_t
 * Job		:	get info about memory handling configuration
 * Notice	:	The I/O mapping is the following:

 	Input char string				output value

	"max_ram"						maximum RAM to allocate
	"max_swap"						maximum SWAP to allocate
	"vm_page_size"					page size for virtual memory
	"total_alloc"					total allocated memory so far
	"total_alloc_ram"				total allocated RAM so far
	"total_alloc_swap"				total allocated SWAP so far

 *--------------------------------------------------------------------------*/

size_t
get_memory_parameter(char * s)
{
	if (!strcmp(s, "max_ram")) {
		return MemoryLimitInMegs ;
	} else if (!strcmp(s, "max_swap")) {
		return VMLimitInMegs ;
	} else if (!strcmp(s, "vm_page_size")) {
		return VMPageSize ;
	} else if (!strcmp(s, "total_alloc")) {
		return (size_t)(total_allocated_RAM + total_allocated_SWAP) ;
	} else if (!strcmp(s, "total_alloc_ram")) {
		return (size_t)total_allocated_RAM ;
	} else if (!strcmp(s, "total_alloc_swap")) {
		return (size_t)total_allocated_SWAP ;
	} else {
		Error("cannot get memory parameter: [%s]\n", s) ;
		return 0 ;
	}
}

/*----------------------------------------------------------------------------
 * Function	:	set_memory_parameter()
 * In 		:	char string, size_t	
 * Out 		:	void
 * Job		:	set info about memory handling configuration
 * Notice	:	The I/O mapping is the following:

 	Input char string				output value

	"max_ram"						maximum RAM to allocate
	"max_swap"						maximum SWAP to allocate
	"vm_page_size"					page size for virtual memory

 *--------------------------------------------------------------------------*/

void
set_memory_parameter(char * s, size_t val)
{
	if (!strcmp(s, "max_ram")) {
		MemoryLimitInMegs = val ;
	} else if (!strcmp(s, "max_swap")) {
		VMLimitInMegs = val ;
	} else if (!strcmp(s, "vm_page_size")) {
		VMPageSize = val ;
	} else {
		Error("cannot set memory parameter: [%s]\n", s) ;
	}
	return ;
}


/*---------------------------------------------------------------------------
 * Function	:	memory_check()
 * In 		:	size (size_t)
 * Out 		:	flag : DISK or MEMORY
 * Job		:	Determines if allocating the requested size will go above
 *				the upper memory limit or not. If it does, the function
 *				returns DISK, which means that data is no loaded into
 *				memory yet. If allocation is possible, it returns MEMORY
 * Notice	:
 *--------------------------------------------------------------------------*/
char	
memory_check(size_t size)
{
	return ((size+total_allocated_RAM)<MemoryLimitInMegs*ONE_MEG)?1:0 ;
}



/*---------------------------------------------------------------------------
 * Function	:	get_tempfile_name()
 * In 		:	void
 * Out 		:	newly allocated character string
 * Job		:	return a different file name at each call
 * Notice	:
 *--------------------------------------------------------------------------*/
char *
get_tempfile_name(void)
{
	char			*name ;

	file_reg++ ;
	name = malloc(NAME_SIZE) ;
	if (name == (char*)NULL) {
		Error("cannot allocate memory to store temporary file name!\n") ;
		return((char*)NULL) ;
	}	
	(void)sprintf(name, "%s/vmswap_%05ld_%05x",
				  tmpdirname,
				  (long)getpid(),
				  (long)file_reg) ;
	return(name) ;
}


/*---------------------------------------------------------------------------
 * Function	:	memory_status()
 * In 		:	void
 * Out 		:	prints out memory status on stderr
 * Notice	:
 *--------------------------------------------------------------------------*/
void	
memory_status(void)
{
	malloc_cell	*curr_cell ;

	if ((total_allocated_RAM!=0) ||
		(total_allocated_SWAP!=0) ||
		(nswapfiles>0) || 
		(nalloc_ptrs>0)) {
		(void)fprintf(stderr, "**** MEMORY STATUS\n") ;
		(void)fprintf(stderr, "memory in use  : %ld bytes\n",
				total_allocated_RAM + total_allocated_SWAP) ;
		(void)fprintf(stderr, "RAM            : %ld bytes\n",
				total_allocated_RAM);
		(void)fprintf(stderr, "SWAP           : %ld bytes\n",
				total_allocated_SWAP);
		(void)fprintf(stderr, "open swapfiles : %d\n", nswapfiles) ;
		
		curr_cell = malloc_table ;
		while (curr_cell != NULL) {

			if (curr_cell->flavour == ALLOC_CELL_RAM) {
				(void)fprintf(stderr, "[RAM] ") ;
			} else {
				(void)fprintf(stderr, "[SWP] ") ;
			}

			(void)fprintf(stderr, "po=%p size=%ld allocated in %s (%d) ",
					curr_cell->pointer,
					curr_cell->size,
					curr_cell->filename,
					curr_cell->lineno) ;

			if (curr_cell->flavour == ALLOC_CELL_SWAP) {
				(void)fprintf(stderr, "in swapfile %s",
						curr_cell->swapfilename) ;
			}
			(void)fprintf(stderr, "\n") ;
			(void)fflush(stderr) ;

			curr_cell = curr_cell->next ;
		}
	}
	return ;
}


/*----------------------------------------------------------------------------
 * Function	:	cleanup_mess()
 * In 		:	void	
 * Out 		:	void	
 * Job		:	clean up temporary files created by the current process
 * Notice	:	
 *--------------------------------------------------------------------------*/

void cleanup_mess(void)
{
	pid_t	pid ;
	char	name[512] ;
	int		i ;

	pid = getpid() ;
	if (pid == (pid_t)-1) {
		Error("cannot get current process ID: no cleanup of tmp files\n") ;
		return ;
	}
	
	for (i=1 ; i<=file_reg ; i++) {
		(void)sprintf(name, "%s/vmswap_%05ld_%05x",
					  tmpdirname,
					  (long)pid,
					  (long)i) ;
		if (remove(name) == 0) {
			fprintf(stderr, "\n\n*** cleaned up tmp file [%s]\n", name) ;
		}
	}
}

void cleanup_mess_wrapper(void)
{ exit(1) ; }


/*----------------------------------------------------------------------------
 * Function	:	interrupt_catch()
 * In 		:	signal to catch, function to call in this case
 * Out 		:	int 0 if worked, -1 otherwise
 * Job		:	catch an interrupt and launch the appropriate function
 * Notice	:	directly taken from MIDAS
 *--------------------------------------------------------------------------*/

int interrupt_catch(void (*f)())
{
	struct sigaction	act,
						oact;
	
	/*
	 * Simply install the given function f() as signal handler for
	 * interrupts
	 */
	act.sa_handler = f;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	if (sigaction(SIGINT,&act,&oact) != 0) return -1;
	return 0 ;
}

#endif
